import { InjectionToken, Injectable, Inject, Directive, TemplateRef, Optional, ViewContainerRef, SkipSelf, Host, Input, NgModule } from '@angular/core';
import { Subject, Observable, of, from } from 'rxjs';
import { map, take, switchMap, takeUntil, startWith, distinctUntilChanged, delay, tap } from 'rxjs/operators';
import { Router } from '@angular/router';

function scanStream(initialValue) {
    const subject = new Subject();
    const obsValues = new Map();
    const values = subject.pipe(map((_) => Array.from(obsValues.values())));
    return {
        values,
        add(obs$) {
            obs$.subscribe((value) => {
                obsValues.set(obs$, value);
                subject.next();
            });
            obsValues.set(obs$, initialValue);
        },
        remove(obs$) {
            obsValues.delete(obs$);
            subject.next();
        },
    };
}

const ACCESS_CONFIG = new InjectionToken('ACCESS_CONFIG');

function flatten(config, { parse = (v) => v, group = false } = {}) {
    const flatConfig = {};
    function setConfig(path, value) {
        if (!flatConfig[path]) {
            flatConfig[path] = new Set();
        }
        flatConfig[path].add(value);
    }
    function getPath(path, delimiter, prop) {
        return path ? path + delimiter + prop : prop;
    }
    function visitor(access, prop, value, path) {
        if (Array.isArray(value)) {
            value.forEach((_access) => {
                visitor(access, prop, _access, path);
            });
        }
        else if (typeof value === 'object') {
            const childrenAccess = children(value, getPath(path, '.', prop));
            if (group) {
                access = access.concat(childrenAccess);
                access.forEach((access) => {
                    setConfig(getPath(path, '.', access.prop), access.action);
                });
            }
        }
        else {
            const expression = parse(value.replace(/\s/g, ''));
            setConfig(getPath(path, '.', prop), expression);
            access = access.concat({ action: expression, prop });
        }
        return access;
    }
    function children(obj, path = '') {
        return Object.keys(obj).reduce((access, prop) => visitor(access, prop, obj[prop], path), []);
    }
    children(config);
    return flatConfig;
}

const TokenType = {
    PAR_OPEN: '('.charCodeAt(0),
    PAR_CLOSE: ')'.charCodeAt(0),
    OP_NOT: '!'.charCodeAt(0),
    BINARY_AND: '&'.charCodeAt(0),
    BINARY_OR: '|'.charCodeAt(0),
    LITERAL: 'LITERAL',
    END: 'END',
    LEAF: 'LEAF',
    ATOMIC: 'ATOMIC',
};

const operator = (evaluator$, op) => (source) => {
    return new Observable((observer) => {
        const finish$ = new Subject();
        let lastEval;
        finish$.pipe(take(1)).subscribe((value) => {
            observer.next(value);
            observer.complete();
        }, (e) => observer.error(e));
        source.pipe(switchMap(evaluator$), takeUntil(finish$)).subscribe((value) => {
            lastEval = op === TokenType.OP_NOT ? !value : value;
            if (op === TokenType.BINARY_AND && !lastEval) {
                finish$.next(false);
            }
            else if (op === TokenType.BINARY_OR && lastEval) {
                finish$.next(true);
            }
        }, (e) => {
            finish$.error(e);
        }, () => {
            finish$.next(lastEval);
        });
    });
};

const Tokenizer = (exp) => {
    let literal = '';
    const tokens = [];
    for (const char of exp) {
        const code = char.charCodeAt(0);
        switch (code) {
            case TokenType.PAR_OPEN:
            case TokenType.PAR_CLOSE:
            case TokenType.OP_NOT:
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (literal) {
                    tokens.push({
                        type: TokenType.LITERAL,
                        value: literal,
                    });
                    literal = '';
                }
                tokens.push({
                    type: code,
                    value: char,
                });
                break;
            default:
                literal += char;
        }
    }
    if (literal)
        tokens.push({
            type: TokenType.LITERAL,
            value: literal,
        });
    return tokens;
};
const ɵ0 = Tokenizer;

const PolishNotation = (tokens) => {
    const stack = [];
    let queue = [];
    let subQueue = [];
    let subStack = [];
    tokens.forEach((token) => {
        switch (token.type) {
            case TokenType.LITERAL:
            case TokenType.OP_NOT:
                if (subStack.length) {
                    subQueue.push(token);
                }
                else {
                    queue.push(token);
                }
                break;
            case TokenType.BINARY_AND:
            case TokenType.BINARY_OR:
                if (subStack.length) {
                    subStack.unshift(token);
                }
                else {
                    stack.unshift(token);
                }
                break;
            case TokenType.PAR_OPEN:
                subStack.unshift(token);
                break;
            case TokenType.PAR_CLOSE:
                while (subStack.length &&
                    subStack[subStack.length - 1].type !== TokenType.PAR_OPEN) {
                    subQueue.unshift(subStack.pop());
                }
                subStack.pop();
                if (subStack.length &&
                    subStack[subStack.length - 1].type === TokenType.OP_NOT) {
                    subQueue.unshift(subStack.pop());
                }
                queue = queue.concat([...subStack, ...subQueue]);
                subStack = [];
                subQueue = [];
                break;
            default:
                break;
        }
    });
    const result = (stack.length && [...stack, ...queue]) || queue;
    return result;
};
const ɵ0$1 = PolishNotation;
const PolishGenerator = function* (polish) {
    for (let index = 0; index < polish.length - 1; index++) {
        yield polish[index];
    }
    return polish[polish.length - 1];
};
const ɵ1 = PolishGenerator;

class ExpNode {
    constructor(op, left, right, literal) {
        this.op = op;
        this.left = left;
        this.right = right;
        this.literal = literal;
    }
    isLeaf() {
        return this.op === TokenType.LEAF;
    }
    isAtomic() {
        var _a;
        return (this.isLeaf() || (this.op === TokenType.OP_NOT && ((_a = this.left) === null || _a === void 0 ? void 0 : _a.isLeaf())));
    }
    getLiteralValue() {
        return this.literal || '';
    }
    static CreateAnd(left, right) {
        return new ExpNode(TokenType.BINARY_AND, left, right);
    }
    static CreateNot(exp) {
        return new ExpNode(TokenType.OP_NOT, exp);
    }
    static CreateOr(left, right) {
        return new ExpNode(TokenType.BINARY_OR, left, right);
    }
    static CreateLiteral(lit) {
        return new ExpNode(TokenType.LEAF, null, null, lit);
    }
}
const make = (gen) => {
    const data = gen.next().value;
    switch (data.type) {
        case TokenType.LITERAL:
            return ExpNode.CreateLiteral(data.value);
        case TokenType.OP_NOT:
            return ExpNode.CreateNot(make(gen));
        case TokenType.BINARY_AND: {
            const left = make(gen);
            const right = make(gen);
            return ExpNode.CreateAnd(left, right);
        }
        case TokenType.BINARY_OR: {
            const left = make(gen);
            const right = make(gen);
            return ExpNode.CreateOr(left, right);
        }
    }
    return null;
};
const nodeEvaluator = (tree, literalEvaluator) => {
    if (tree) {
        if (tree.isLeaf()) {
            return literalEvaluator(tree.getLiteralValue());
        }
        if (tree.op === TokenType.OP_NOT) {
            return !nodeEvaluator(tree.left, literalEvaluator);
        }
        if (tree.op === TokenType.BINARY_OR) {
            return (nodeEvaluator(tree.left, literalEvaluator) ||
                nodeEvaluator(tree.right, literalEvaluator));
        }
        if (tree.op === TokenType.BINARY_AND) {
            return (nodeEvaluator(tree.left, literalEvaluator) &&
                nodeEvaluator(tree.right, literalEvaluator));
        }
    }
    return false;
};

const parser = (exp) => {
    const tokens = Tokenizer(exp.replace(/\s/g, ''));
    const polish = PolishNotation(tokens);
    const gen = PolishGenerator(polish);
    const tree = make(gen);
    return tree;
};
const ɵ0$2 = parser;

let configurationAccess = {};
let hasAccessStrategy = (accessName) => of(false);
const ɵ0$3 = hasAccessStrategy;
function setConfigurationAccess(config) {
    configurationAccess = flatten(config, { parse: parser, group: true });
}
function setHasAccessStrategy(accessTest) {
    hasAccessStrategy = accessTest;
}
function canAccessExpression(accessExpression) {
    return of(accessExpression).pipe(map((ae) => ae.replace(/\s/g, '')), switchMap((ae) => nodeEvaluator$1(parser(ae), hasAccessStrategy)));
}
function canAccessPaths(accessExpression) {
    return of(accessExpression).pipe(map((ae) => ae.replace(/\s/g, '')), switchMap((ae) => nodeEvaluator$1(parser(ae), (path) => canAccessPath(path))));
}
function canAccessPath(path) {
    try {
        const access = configurationAccess[path];
        if (!!access) {
            return from(access).pipe(operator((tree) => nodeEvaluator$1(tree, hasAccessStrategy), TokenType.BINARY_OR));
        }
        console.error(`Undefined path ${path}`);
        return of(false);
    }
    catch (e) {
        console.error(e);
        return of(false);
    }
}
function nodeEvaluator$1(tree, literalEvaluator) {
    if (tree) {
        if (tree.isLeaf()) {
            const check = literalEvaluator(tree.getLiteralValue());
            return typeof check === 'boolean' ? of(check) : check;
        }
        if (tree.op === TokenType.OP_NOT) {
            return of(tree.left).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.OP_NOT));
        }
        if (tree.op === TokenType.BINARY_OR) {
            return from([tree.left, tree.right]).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.BINARY_OR));
        }
        if (tree.op === TokenType.BINARY_AND) {
            return from([tree.left, tree.right]).pipe(operator((tree) => nodeEvaluator$1(tree, literalEvaluator), TokenType.BINARY_AND));
        }
    }
    return of(false);
}

class AccessStrategy {
}

class AccessService {
    constructor(config, accessStrategy) {
        setConfigurationAccess(config.access || {});
        setHasAccessStrategy((accessName) => accessStrategy.has(accessName));
    }
    setConfiguration(access = {}) {
        setConfigurationAccess(access);
    }
    setAccessStrategy(accessStrategy) {
        setHasAccessStrategy((accessName) => accessStrategy.has(accessName));
    }
    can(access) {
        return access.startsWith(':')
            ? canAccessPaths(access.substring(1))
            : canAccessExpression(access);
    }
}
AccessService.decorators = [
    { type: Injectable }
];
AccessService.ctorParameters = () => [
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] },
    { type: AccessStrategy }
];

class AccessDirective {
    constructor(template, viewContainer, accessService, parentAccessDirective) {
        this.template = template;
        this.viewContainer = viewContainer;
        this.accessService = accessService;
        this.parentAccessDirective = parentAccessDirective;
        this.ngxAccess = '';
        this.onDestroy$ = new Subject();
        this._access$ = new Subject();
        this.access$ = this._access$.asObservable();
    }
    ngOnInit() {
        if (this.template) {
            if (this.ngxAccess) {
                let ngxAccess = this.ngxAccess;
                // if (
                //   this.parentAccessDirective &&
                //   this.parentAccessDirective.ngxAccess
                // ) {
                //   const { path, action } = parse(this.parentAccessDirective.ngxAccess);
                //   if (ngxAccess) {
                //     if (Array.isArray(this.ngxAccess)) {
                //       ngxAccess = `${this.ngxAccess.map(
                //         (access) => path + access
                //       )}:${action}`;
                //     } else {
                //       const { path: childPath, action: childAction } = parse(ngxAccess);
                //       ngxAccess = `${path}${childPath}:${
                //         childAction ? childAction : action
                //       }`;
                //     }
                //   } else {
                //     ngxAccess = `${path}:${action}`;
                //   }
                // }
                this.check(ngxAccess);
            }
            else {
                this.children$ = scanStream(true);
                this.children$.values
                    .pipe(takeUntil(this.onDestroy$), map((values) => {
                    return values.reduce((acc, val) => {
                        return acc || val;
                    }, false);
                }), startWith(true), distinctUntilChanged())
                    .subscribe((access) => {
                    if (access) {
                        this.viewContainer.createEmbeddedView(this.template);
                    }
                    else {
                        this.viewContainer.clear();
                    }
                });
            }
        }
    }
    observe(child) {
        var _a;
        (_a = this.children$) === null || _a === void 0 ? void 0 : _a.add(child);
    }
    unObserve(child) {
        var _a;
        (_a = this.children$) === null || _a === void 0 ? void 0 : _a.remove(child);
    }
    check(access) {
        this.access$ = this.accessService
            .can(access)
            .pipe(takeUntil(this.onDestroy$));
        if (this.parentAccessDirective) {
            this.parentAccessDirective.observe(this.access$.pipe(map((access) => access || !!this.ngxAccessElse), delay(0)));
        }
        this.access$.subscribe((access) => {
            return access
                ? this.viewContainer.createEmbeddedView(this.template)
                : this.ngxAccessElse
                    ? this.viewContainer.createEmbeddedView(this.ngxAccessElse)
                    : null;
        });
    }
    ngOnDestroy() {
        if (this.parentAccessDirective) {
            this.parentAccessDirective.unObserve(this.access$);
        }
        this.onDestroy$.next();
        this.onDestroy$.complete();
    }
}
AccessDirective.decorators = [
    { type: Directive, args: [{
                selector: '[ngxAccess]',
            },] }
];
AccessDirective.ctorParameters = () => [
    { type: TemplateRef, decorators: [{ type: Optional }] },
    { type: ViewContainerRef },
    { type: AccessService },
    { type: AccessDirective, decorators: [{ type: Optional }, { type: SkipSelf }, { type: Host }] }
];
AccessDirective.propDecorators = {
    ngxAccess: [{ type: Input }],
    ngxAccessElse: [{ type: Input }]
};

class AccessGuard {
    constructor(router, accessService, config) {
        this.router = router;
        this.accessService = accessService;
        this.config = config;
    }
    canActivate(next) {
        return this.accessService
            .can(next.data.access)
            .pipe(tap((hasAccess) => !hasAccess &&
            this.router.navigate([
                next.data.redirectTo || this.config.redirectTo,
            ])));
    }
    canLoad(route) {
        var _a;
        return this.accessService
            .can((_a = route.data) === null || _a === void 0 ? void 0 : _a.access)
            .pipe(tap((hasAccess) => {
            var _a;
            return !hasAccess &&
                this.router.navigate([
                    ((_a = route.data) === null || _a === void 0 ? void 0 : _a.redirectTo) || this.config.redirectTo,
                ]);
        }));
    }
}
AccessGuard.decorators = [
    { type: Injectable }
];
AccessGuard.ctorParameters = () => [
    { type: Router },
    { type: AccessService },
    { type: undefined, decorators: [{ type: Inject, args: [ACCESS_CONFIG,] }] }
];

class AccessModule {
    static forRoot(config) {
        return {
            ngModule: AccessModule,
            providers: [
                AccessService,
                AccessGuard,
                {
                    provide: ACCESS_CONFIG,
                    useValue: {
                        access: config.access || {},
                        redirectTo: config.redirectTo || '/unauthorized',
                    },
                },
            ],
        };
    }
}
AccessModule.decorators = [
    { type: NgModule, args: [{
                declarations: [AccessDirective],
                exports: [AccessDirective],
            },] }
];

/**
 * Generated bundle index. Do not edit.
 */

export { ACCESS_CONFIG, AccessDirective, AccessGuard, AccessModule, AccessService, AccessStrategy, AccessStrategy as ɵa, AccessDirective as ɵb };
//# sourceMappingURL=ngx-access.js.map
